import numpy as np
import math
import torch
from tqdm import tqdm
import warnings

SQRTEPS = math.sqrt(float(np.finfo(np.float64).eps))


def trim_weights(samples, weights, ess=0.99, bins=1000):
    """
        Trim samples and weights to a given effective sample size.

    Parameters
    ----------
    samples : ``np.ndarray``
        Samples.
    weights : ``np.ndarray``
        Weights.
    ess : ``float``
        Effective sample size threshold.
    bins : ``int``
        Number of bins to use for trimming.

    Returns
    -------
    samples_trimmed : ``np.ndarray``
        Trimmed samples.
    weights_trimmed : ``np.ndarray``
        Trimmed weights.
    """

    # normalize weights
    weights /= np.sum(weights)
    # compute untrimmed ess
    ess_total = 1.0 / np.sum(weights**2.0)
    # define percentile grid
    percentiles = np.linspace(0, 99, bins)

    i = bins - 1
    while True:
        p = percentiles[i]
        # compute weight threshold
        threshold = np.percentile(weights, p)
        mask = weights >= threshold
        weights_trimmed = weights[mask]
        weights_trimmed /= np.sum(weights_trimmed)
        ess_trimmed = 1.0 / np.sum(weights_trimmed**2.0)
        if ess_trimmed / ess_total >= ess:
            break 
        i -= 1
    
    return samples[mask], weights_trimmed


def effective_sample_size(weights):
    """
        Compute effective sample size (ESS).

    Parameters
    ----------
    weights : ``np.ndarray``
        Weights.

    Returns
    -------
    ess : ``float``
        Effective sample size.
    """
    weights /= np.sum(weights)
    return 1.0 / np.sum(weights**2.0)


def unique_sample_size(weights, k=None):
    """
        Compute unique sample size (ESS).

    Parameters
    ----------
    weights : ``np.ndarray``
        Weights.
    k : ``int``
        Number of resampled samples.

    Returns
    -------
    uss : ``float``
        Unique sample size.
    """
    if k is None:
        k = len(weights)
    weights /= np.sum(weights)
    return np.sum(1.0 - (1.0 - weights)**k)


def compute_ess(logw: np.ndarray):
    r"""
        Compute effective sample size (per centage).

    Parameters
    ----------
    logw : ``np.ndarray``
        Log-weights.
    Returns
    -------
    ess : float
        Effective sample size divided by actual number
        of particles (between 0 and 1)
    """
    logw_max = np.max(logw)
    logw_normed = logw - logw_max

    weights = np.exp(logw_normed) / np.sum(np.exp(logw_normed))
    return 1.0 / np.sum(weights * weights) / len(weights)


def increment_logz(logw: np.ndarray):
    r"""
        Compute log evidence increment factor.

    Parameters
    ----------
    logw : ``np.ndarray``
        Log-weights.
    Returns
    -------
    ess : float
        logZ increment.
    """
    logw_max = np.max(logw)
    logw_normed = logw - logw_max

    return logw_max + np.logaddexp.reduce(logw_normed)


def systematic_resample(size: np.ndarray, 
                        weights: np.ndarray, 
                        random_state: int = None):
    """
        Resample a new set of points from the weighted set of inputs
        such that they all have equal weight.

    Parameters
    ----------
    size : `int`
        Number of samples to draw.
    weights : `~numpy.ndarray` with shape (nsamples,)
        Corresponding weight of each sample.
    random_state : `int`, optional
        Random seed.    

    Returns
    -------
    indeces : `~numpy.ndarray` with shape (nsamples,)
        Indices of the resampled array.
    
    Examples
    --------
    >>> x = np.array([[1., 1.], [2., 2.], [3., 3.], [4., 4.]])
    >>> w = np.array([0.6, 0.2, 0.15, 0.05])
    >>> systematic_resample(4, w)
    array([0, 0, 0, 2])

    Notes
    -----
    Implements the systematic resampling method.
    """
    
    if random_state is not None:
        np.random.seed(random_state)

    if abs(np.sum(weights) - 1.) > SQRTEPS:
        weights = np.array(weights) / np.sum(weights)

    positions = (np.random.random() + np.arange(size)) / size

    j = 0
    cumulative_sum = weights[0]
    indeces = np.empty(size, dtype=int)
    for i in range(size):
        while positions[i] > cumulative_sum:
            j += 1
            cumulative_sum += weights[j]
        indeces[i] = j
    
    return indeces


class ProgressBar:
    """
        Progress bar class.

    Parameters
    ----------
    show : `bool`
        Whether or not to show a progress bar. Default is ``True``.
    """
    def __init__(self, show: bool = True):
        self.progress_bar = tqdm(desc='Iter', disable=not show)
        self.info = dict()

    def update_stats(self, info):
        """
            Update shown stats.

        Parameters
        ----------
        info : dict
            Dictionary with stats to show.
        """
        self.info = {**self.info, **info}
        self.progress_bar.set_postfix(ordered_dict=self.info)

    def update_iter(self):
        """
            Update iteration counter.
        """
        self.progress_bar.update(1)

    def close(self):
        """
            Close progress bar.
        """
        self.progress_bar.close()


class FunctionWrapper(object):
    r"""
        Make the log-likelihood or log-prior function pickleable
        when ``args`` or ``kwargs`` are also included.

    Parameters
    ----------
    f : callable
        Log probability function.
    args : list
        Extra positional arguments to be passed to f.
    kwargs : dict
        Extra keyword arguments to be passed to f.
    """
    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = [] if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

    def __call__(self, x):
        """
            Evaluate log-likelihood or log-prior function.

        Parameters
        ----------
        x : ``np.ndarray``
            Input position array.

        Returns
        -------
        f : float or ``np.ndarray``
            f(x)
        """
        return self.f(x, *self.args, **self.kwargs)


def torch_to_numpy(x: torch.Tensor) -> np.ndarray:
    """
    Cast torch tensor to numpy.

    Parameters
    ----------
    x : torch.Tensor
        Input tensor.

    Returns
    -------
        Numpy array corresponding to the input tensor.
    """
    return x.detach().numpy()


def numpy_to_torch(x: np.ndarray) -> torch.Tensor:
    """
    Cast numpy array to torch tensor.

    Parameters
    ----------
    x : np.ndarray
        Input array.

    Returns
    -------
        Torch tensor corresponding to the input array.
    """
    return torch.tensor(x, dtype=torch.float32)


def torch_double_to_float(x: torch.Tensor, warn: bool = True):
    """
    Cast double precision (Float64) torch tensor to single precision (Float32).

    Parameters
    ----------
    x: torch.Tensor
        Input tensor.
    warn: bool
        If True, warn the user about the typecast.

    Returns
    -------
        Single precision (Float32) torch tensor.
    """
    if x.dtype == torch.float64 and warn:
        warnings.warn(f"Float64 data is currently unsupported, casting to Float32. Output will also have type Float32.")
        return x.float()
    elif x.dtype == torch.float32:
        return x
    else:
        raise ValueError(f"Unsupported datatype for input data: {x.dtype}")

class flow_numpy_wrapper:
    """
    Wrapper class for numpy flows.

    Parameters
    ----------
    flow : Flow object
        Flow object that implements forward and inverse
        transformations.
    
    Returns
    -------
    Flow object
    """
    def __init__(self, flow):
        self.flow = flow

    @torch.no_grad()
    def forward(self, v):
        v = numpy_to_torch(v)
        theta, logdetj = self.flow.forward(v)
        theta = torch_to_numpy(theta)
        logdetj = - torch_to_numpy(logdetj)
        return theta, logdetj

    @torch.no_grad()
    def inverse(self, theta):
        theta = numpy_to_torch(theta)
        v, logdetj = self.flow.inverse(theta)
        v = torch_to_numpy(v)
        logdetj = torch_to_numpy(logdetj)
        return v, logdetj


from typing import Union, Optional, Tuple, Dict, List

import numpy as np
import copy
import time 
import zuko
import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.optim.lr_scheduler import ReduceLROnPlateau

from .tools import torch_double_to_float

class Flow:
    """
    Normalizing flow model.

    Parameters
    ----------
    n_dim : ``int``
        Number of dimensions of the distribution to be modeled.
    flow : ``zuko.flows.Flow`` or str, optional
        Normalizing flow model. Default: ``nsf3``.

    Attributes
    ----------
    n_dim : ``int``
        Number of dimensions of the distribution to be modeled.
    flow : ``zuko.flows.Flow``
        Normalizing flow model.
    transform : ``zuko.transforms.Transform``
        Transformation object.
    
    Examples
    --------
    >>> import torch
    >>> import pocomc
    >>> flow = pocomc.Flow(2)
    >>> x = torch.randn(100, 2)
    >>> u, logdetj = flow(x)
    >>> x_, logdetj_ = flow.inverse(u)
    >>> log_prob = flow.log_prob(x)
    >>> x_, log_prob_ = flow.sample(100)
    >>> history = flow.fit(x)
    """

    def __init__(self, n_dim, flow='nsf3'):
        self.n_dim = n_dim

        def next_power_of_2(n):
            return 1 if n == 0 else 2**(n - 1).bit_length()
        
        n_hidden = np.maximum(next_power_of_2(3*n_dim), 32)

        if flow == 'maf3':
            self.flow = zuko.flows.MAF(n_dim, 
                                       transforms=3, 
                                       hidden_features=[n_hidden] * 3,
                                       residual=True,)
        elif flow == 'maf6':
            self.flow = zuko.flows.MAF(n_dim, 
                                       transforms=6, 
                                       hidden_features=[n_hidden] * 3,
                                       residual=True,)
        elif flow == 'maf12':
            self.flow = zuko.flows.MAF(n_dim, 
                                       transforms=12, 
                                       hidden_features=[n_hidden] * 3,
                                       residual=True,)
        elif flow == 'nsf3':
            self.flow = zuko.flows.NSF(features=n_dim, 
                                       bins=8, 
                                       transforms=3, 
                                       hidden_features=[n_hidden] * 3,
                                       residual=True)
        elif flow == 'nsf6':
            self.flow = zuko.flows.NSF(features=n_dim, 
                                       bins=8, 
                                       transforms=6, 
                                       hidden_features=[n_hidden] * 3,
                                       residual=True)
        elif flow == 'nsf12':
            self.flow = zuko.flows.NSF(features=n_dim, 
                                       bins=8, 
                                       transforms=12, 
                                       hidden_features=[n_hidden] * 3,
                                       residual=True)
        elif isinstance(flow, zuko.flows.Flow):
            self.flow = flow
        else:
            raise ValueError('Invalid flow type. Choose from: maf3, maf6, maf12, nsf3, nsf6, nsf12, or provide a zuko.flows.Flow object.')

    @property
    def transform(self):
        """
        Transformation object.
        """
        return self.flow().transform

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Forward transformation.
        Inputs are transformed from the original (relating to the distribution to be modeled) to the latent space.

        Parameters
        ----------
        x : ``torch.Tensor``
            Samples to transform.
        Returns
        -------
        u : ``torch.Tensor``
            Transformed samples in latent space with the same shape as the original space inputs.
        """
        x = torch_double_to_float(x)
        return self.transform.call_and_ladj(x)

    def inverse(self, u: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Inverse transformation.
        Inputs are transformed from the latent to the original space (relating to the distribution to be modeled).
        
        Parameters
        ----------
        u : ``torch.Tensor``
            Samples to transform.
        Returns
        -------
        x : ``torch.Tensor``
            Transformed samples in the original space with the same shape as the latent space inputs.
        """
        u = torch_double_to_float(u)
        x, logdetj = self.transform.inv.call_and_ladj(u)
        return x, logdetj

    def log_prob(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute log probability of samples.
        
        Parameters
        ----------
        x : ``torch.Tensor``
            Input samples
        Returns
        -------
        Log-probability of samples.
        """
        x = torch_double_to_float(x)
        return self.flow().log_prob(x)

    def sample(self, size: int = 1) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Draw random samples from the normalizing flow.

        Parameters
        ----------
        size : ``int``
            Number of samples to generate. Default: 1.
        Returns
        -------
        samples, log_prob : ``tuple``
            Samples as a ``torch.Tensor`` with shape ``(size, n_dimensions)`` and log probability values with shape ``(size, )``.
        """
        x, log_p = self.flow().rsample_and_log_prob((size,))
        return x, log_p

    def fit(self,
            x,
            weights=None,
            validation_split=0.0,
            epochs=1000,
            batch_size=1000,
            patience=20,
            learning_rate=1e-3,
            weight_decay=0,
            laplace_scale=None,
            gaussian_scale=None,
            annealing=True,
            noise=None,
            shuffle=True,
            clip_grad_norm=1.0,
            verbose=0,
            ):
        """

        Parameters
        ----------
        x : ``torch.Tensor``
            Input samples.
        weights : ``torch.Tensor``, optional
            Weights for each sample. Default: ``None``.
        validation_split : ``float``, optional
            Fraction of samples to use for validation. Default: 0.0.
        epochs : ``int``, optional
            Number of epochs. Default: 1000.
        batch_size : ``int``, optional
            Batch size. Default: 1000.
        patience : ``int``, optional
            Number of epochs without improvement before early stopping. Default: 20.
        learning_rate : ``float``, optional
            Learning rate. Default: 1e-3.
        weight_decay : ``float``, optional
            Weight decay. Default: 0.
        laplace_scale : ``float``, optional
            Laplace regularization scale. Default: ``None``.
        gaussian_scale : ``float``, optional
            Gaussian regularization scale. Default: ``None``.
        annealing : ``bool``, optional
            Whether to use learning rate annealing. Default: ``True``.
        noise : ``float``, optional
            Noise scale. Default: ``None``.
        shuffle : ``bool``, optional
            Whether to shuffle samples. Default: ``True``.
        clip_grad_norm : ``float``, optional
            Maximum gradient norm. Default: 1.0.
        verbose : ``int``, optional
            Verbosity level. Default: 0.

        Returns
        -------
        history : ``dict``
            Dictionary with loss history.

        Examples
        --------
        >>> import torch
        >>> import pocomc
        >>> flow = pocomc.Flow(2)
        >>> x = torch.randn(100, 2)
        >>> history = flow.fit(x)    
        """
        x = torch_double_to_float(x)

        n_samples, n_dim = x.shape

        if shuffle:
            rand_indx = torch.randperm(n_samples)
            x = x[rand_indx]
            if weights is not None:
                weights = weights[rand_indx]

        if noise is not None:
            min_dists = torch.empty(n_samples)
            for i in range(n_samples):
                min_dist = torch.linalg.norm(x[i] - x, axis=1)
                min_dists[i] = torch.min(min_dist[min_dist > 0.0])
            mean_min_dist = torch.mean(min_dist)

        if validation_split > 0.0:
            x_train = x[:int(validation_split * n_samples)]
            x_valid = x[int(validation_split * n_samples):]
            if weights is None:
                train_dl = DataLoader(TensorDataset(x_train), batch_size, shuffle)
                val_dl = DataLoader(TensorDataset(x_valid), batch_size, shuffle)
            else:
                weights_train = weights[:int(validation_split * n_samples)]
                weights_valid = weights[int(validation_split * n_samples):]
                train_dl = DataLoader(TensorDataset(x_train, weights_train), batch_size, shuffle)
                val_dl = DataLoader(TensorDataset(x_valid, weights_valid), batch_size, shuffle)
            validation = True
        else:
            x_train = x
            if weights is None:
                train_dl = DataLoader(TensorDataset(x_train), batch_size, shuffle)
            else:
                weights_train = weights
                train_dl = DataLoader(TensorDataset(x_train, weights_train), batch_size, shuffle)
            validation = False

        optimizer = torch.optim.AdamW(self.flow.parameters(), 
                                      learning_rate,
                                      weight_decay=weight_decay,
                                      )

        if annealing:
            scheduler = ReduceLROnPlateau(optimizer, 
                                          mode='min',
                                          factor=0.2,
                                          patience=patience, 
                                          threshold=0.0001, 
                                          threshold_mode='abs', 
                                          min_lr=1e-6,
                                         )

        history = dict()  # Collects per-epoch loss
        history['loss'] = []
        history['val_loss'] = []
        if validation:
            monitor = 'val_loss'
        else:
            monitor = 'loss'

        best_epoch = 0
        best_loss = np.inf
        best_model = copy.deepcopy(self.flow.state_dict())

        start_time_sec = time.time()

        for epoch in range(epochs):
            self.flow.train()
            train_loss = 0.0

            for batch in train_dl:

                optimizer.zero_grad()
                if noise is None:
                    x_ = batch[0]
                else:
                    x_ = batch[0] + noise * mean_min_dist * torch.randn_like(batch[0])
                if weights is None:
                    loss = -self.flow().log_prob(x_).sum()
                else:
                    loss = -self.flow().log_prob(x_) * batch[1] * 1000.0
                    loss = loss.sum() / batch[1].sum()

                if laplace_scale is not None or gaussian_scale is not None:
                    loss -= regularization_loss(self.flow, laplace_scale, gaussian_scale)

                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.flow.parameters(), clip_grad_norm)
                optimizer.step()

                train_loss += loss.data.item()

            train_loss = train_loss / len(train_dl.dataset)

            history['loss'].append(train_loss)
            
            if validation:
                self.flow.eval()
                val_loss = 0.0

                for batch in val_dl:

                    if noise is None:
                        x_ = batch[0]
                    else:
                        x_ = batch[0] + noise * mean_min_dist * torch.randn_like(batch[0])
                    if weights is None:
                        loss = -self.flow().log_prob(x_).sum()
                    else:
                        loss = -self.flow().log_prob(x_) * batch[1] * 1000.0
                        loss = loss.sum() / batch[1].sum()
                    
                    if laplace_scale is not None or gaussian_scale is not None:
                        loss -= regularization_loss(self.flow, laplace_scale, gaussian_scale)

                    val_loss += loss.data.item()

                val_loss = val_loss / len(val_dl.dataset)

                history['val_loss'].append(val_loss)
        
            if annealing and validation:
                scheduler.step(val_loss)
            elif annealing:
                scheduler.step(train_loss)

            if verbose > 1:
                try:
                    print('Epoch %3d/%3d, train loss: %5.2f, val loss: %5.2f' % (epoch + 1, epochs, train_loss, val_loss))
                except:  # TODO specify type of exception
                    print('Epoch %3d/%3d, train loss: %5.2f' % (epoch + 1, epochs, train_loss))

            # Monitor loss
            if history[monitor][-1] < best_loss:
                best_loss = history[monitor][-1]
                best_epoch = epoch
                best_model = copy.deepcopy(self.flow.state_dict())

            if epoch - best_epoch >= int(1.5 * patience):
                self.flow.load_state_dict(best_model)
                if verbose > 0:
                    print('Finished early after %3d epochs' % best_epoch)
                    print('Best loss achieved %5.2f' % best_loss)
                break
        
        if verbose > 0:
            end_time_sec = time.time()
            total_time_sec = end_time_sec - start_time_sec
            time_per_epoch_sec = total_time_sec / epochs
            print()
            print('Time total:     %5.2f sec' % total_time_sec)
            print('Time per epoch: %5.2f sec' % time_per_epoch_sec)

        return history


def regularization_loss(model, laplace_scale=None, gaussian_scale=None):
    """
    Compute regularization loss.

    Parameters
    ----------
    model : ``zuko.flows.Flow``
        Normalizing flow model.
    laplace_scale : ``float``, optional
        Laplace regularization scale. Default: ``None``.
    gaussian_scale : ``float``, optional
        Gaussian regularization scale. Default: ``None``.
    
    Returns
    -------
    Regularization loss.
    """
    total_laplace = 0.0
    total_gaussian = 0.0

    for i, transform in enumerate(model.transforms):
        if hasattr(transform, "hyper"):
            for parameter_name, parameter in transform.hyper.named_parameters():
                if parameter_name.endswith('weight'):
                    if laplace_scale is not None:
                        total_laplace += parameter.abs().sum()
                    if gaussian_scale is not None:
                        total_gaussian += parameter.square().sum()
    
    total = 0.0
    if laplace_scale is not None:
        total += - total_laplace / laplace_scale
    if gaussian_scale is not None:
        total += - total_gaussian / (2.0 * gaussian_scale**2.0)

    return total


import numpy as np

from .student import fit_mvstud
from .tools import systematic_resample

class Geometry:
    """
    Geometry class for the POCOMC algorithm.

    Attributes
    ----------
    normal_mean : array_like
        Mean of the normal distribution.
    normal_cov : array_like
        Covariance matrix of the normal distribution.
    t_mean : array_like
        Mean of the t distribution.
    t_cov : array_like
        Covariance matrix of the t distribution.
    t_nu : float
        Degrees of freedom of the t distribution.
    """

    def __init__(self):
        self.normal_mean = None
        self.normal_cov = None
        self.t_mean = None
        self.t_cov = None
        self.t_nu = None

    def fit(self, theta, weights=None):
        """

        Parameters
        ----------
        theta : array_like
            Array of samples.
        weights : array_like, optional
            Array of weights. The default is None.
        """
        
        # Learn normal distribution
        if weights is None:
            self.normal_mean = np.mean(theta, axis=0)
            self.normal_cov = np.cov(theta.T)
        else:
            self.normal_mean = np.average(theta, axis=0, weights=weights)
            self.normal_cov = np.cov(theta.T, aweights=weights)

        # Learn t distribution
        if weights is not None:
            idx_resampled = systematic_resample(len(theta), weights=weights)
            theta_resampled = theta[idx_resampled]
            self.t_mean, self.t_cov, self.t_nu = fit_mvstud(theta_resampled)
        else:
            self.t_mean, self.t_cov, self.t_nu = fit_mvstud(theta)

        if ~np.isfinite(self.t_nu):
            self.t_nu = 1e6


import unittest

import numpy as np
from scipy.stats import norm

from pocomc.prior import Prior

class PriorTestCase(unittest.TestCase):

    def test_sample(self):
        prior = Prior([norm(0, 1), norm(0, 1)])
        x = prior.rvs(10)
        self.assertEqual(np.shape(x), (10,2))

    def test_log_prob(self):
        prior = Prior([norm(0, 1), norm(0, 1)])
        x = prior.rvs(10)
        log_prob = prior.logpdf(x)
        self.assertIsInstance(log_prob, np.ndarray)

    def test_log_prob2(self):
        prior = Prior([norm(0, 1), norm(0, 1)])
        x = prior.rvs(10)
        log_prob = prior.logpdf(x)
        self.assertEqual(np.shape(log_prob), (10,))

    def test_log_prob3(self):
        prior = Prior([norm(0, 1), norm(0, 1)])
        x = prior.rvs(10)
        log_prob = prior.logpdf(x)
        self.assertTrue(np.all(log_prob < 0))
    
    def test_log_prob4(self):
        prior = Prior([norm(0, 1), norm(0, 1)])
        x = prior.rvs(10)
        log_prob = prior.logpdf(x)
        self.assertTrue(np.all(np.isfinite(log_prob)))

    def test_bounds(self):
        prior = Prior([norm(0, 1), norm(0, 1)])
        bounds = prior.bounds
        self.assertEqual(np.shape(bounds), (2,2))

    def test_bounds2(self):
        prior = Prior([norm(0, 1), norm(0, 1)])
        bounds = prior.bounds
        self.assertTrue(np.all(bounds[:,0] < bounds[:,1]))

    def test_dim(self):
        prior = Prior([norm(0, 1), norm(0, 1)])
        self.assertEqual(prior.dim, 2)


    


import unittest
import numpy as np

from pocomc.scaler import Reparameterize


class ReparameterizeTestCase(unittest.TestCase):
    @staticmethod
    def make_unconstrained_data():
        # Make a dataset to use in tests
        np.random.seed(0)
        n_data = 100
        n_dim = 10
        x = np.random.randn(n_data, n_dim) * 3 + 1

        lower_bound = np.nan
        upper_bound = np.nan
        return x, lower_bound, upper_bound

    @staticmethod
    def make_lower_bounded_data():
        # Make a dataset to use in tests
        np.random.seed(0)
        n_data = 100
        n_dim = 10
        x = np.random.exponential(scale=1, size=(n_data, n_dim))

        lower_bound = 0
        upper_bound = np.nan
        return x, lower_bound, upper_bound

    @staticmethod
    def make_upper_bounded_data():
        # Make a dataset to use in tests
        np.random.seed(0)
        n_data = 100
        n_dim = 10
        x = -np.random.exponential(scale=1, size=(n_data, n_dim))

        lower_bound = np.nan
        upper_bound = 0
        return x, lower_bound, upper_bound

    @staticmethod
    def make_lower_and_upper_bounded_data():
        # Make a dataset to use in tests
        np.random.seed(0)
        n_data = 100
        n_dim = 10
        x = np.random.uniform(low=0, high=1, size=(n_data, n_dim))

        lower_bound = 0
        upper_bound = 1
        return x, lower_bound, upper_bound

    def test_unconstrained(self):
        # Test that methods work without errors on unconstrained input data
        np.random.seed(0)

        x, lb, ub = self.make_unconstrained_data()
        r = Reparameterize(n_dim=x.shape[1], bounds=(lb, ub))
        r.fit(x)

        u = r.forward(x)
        x_r, _ = r.inverse(u)

        self.assertEqual(x.shape, u.shape)
        self.assertEqual(x.dtype, u.dtype)

        self.assertEqual(x_r.shape, u.shape)
        self.assertEqual(x_r.dtype, u.dtype)

        self.assertTrue(np.allclose(x, x_r))

    def test_lower_bounded(self):
        # Test that methods work without errors on unconstrained input data
        np.random.seed(0)

        x, lb, ub = self.make_lower_bounded_data()
        r = Reparameterize(n_dim=x.shape[1], bounds=(lb, ub))
        r.fit(x)

        u = r.forward(x)
        x_r, _ = r.inverse(u)

        self.assertEqual(x.shape, u.shape)
        self.assertEqual(x.dtype, u.dtype)

        self.assertEqual(x_r.shape, u.shape)
        self.assertEqual(x_r.dtype, u.dtype)

        self.assertTrue(np.allclose(x, x_r))

    def test_upper_bounded(self):
        # Test that methods work without errors on unconstrained input data
        np.random.seed(0)

        x, lb, ub = self.make_upper_bounded_data()
        r = Reparameterize(n_dim=x.shape[1], bounds=(lb, ub))
        r.fit(x)

        u = r.forward(x)
        x_r, _ = r.inverse(u)

        self.assertEqual(x.shape, u.shape)
        self.assertEqual(x.dtype, u.dtype)

        self.assertEqual(x_r.shape, u.shape)
        self.assertEqual(x_r.dtype, u.dtype)

        self.assertTrue(np.allclose(x, x_r))

    def test_lower_and_upper_bounded(self):
        # Test that methods work without errors on unconstrained input data
        np.random.seed(0)

        x, lb, ub = self.make_lower_and_upper_bounded_data()
        r = Reparameterize(n_dim=x.shape[1], bounds=(lb, ub))
        r.fit(x)

        u = r.forward(x)
        x_r, _ = r.inverse(u)

        self.assertEqual(x.shape, u.shape)
        self.assertEqual(x.dtype, u.dtype)

        self.assertEqual(x_r.shape, u.shape)
        self.assertEqual(x_r.dtype, u.dtype)

        self.assertTrue(np.allclose(x, x_r))

    def test_out_of_bounds(self):
        # Test that providing out-of-bound inputs raises an error
        np.random.seed(0)
        x, lb, ub = self.make_lower_and_upper_bounded_data()
        x[0] = lb - 1  # Artificially make example 0 go outside the bounds
        x[1] = ub + 1  # Artificially make example 1 go outside the bounds

        r = Reparameterize(n_dim=x.shape[1], bounds=(lb, ub))
        self.assertRaises(ValueError, r.fit, x)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from scipy import optimize
from scipy import special

def fit_mvstud(data, tolerance=1e-6, max_iter=100):
    """
    Fit a multivariate Student's t distribution to data using the EM algorithm.

    Parameters
    ----------
    data : ndarray
        An array of shape (dim, n) containing n samples of dimension dim.
    tolerance : float, optional
        The tolerance for convergence. The default is 1e-6.
    max_iter : int, optional
        The maximum number of iterations. The default is 100.
    
    Returns
    -------
    mu : ndarray
        The mean of the distribution.
    Sigma : ndarray
        The covariance matrix of the distribution.
    nu : float
        The degrees of freedom of the distribution.
    
    Examples
    --------
    >>> import numpy as np
    >>> from pocomc.student import fit_mvstud
    >>> data = np.random.randn(2, 100)
    >>> mu, Sigma, nu = fit_mvstud(data)
    >>> mu
    array([ 0.00323705, -0.05405479])
    >>> Sigma
    array([[ 1.00524016, -0.02020086],
           [-0.02020086,  0.99111344]])
    >>> nu
    20.000000000000004
    """
    def opt_nu(delta_iobs, nu):
        def func0(nu):
            w_iobs = (nu + dim) / (nu + delta_iobs)
            f = -special.psi(nu/2) + np.log(nu/2) + np.sum(np.log(w_iobs))/n - np.sum(w_iobs)/n + 1 + special.psi((nu+dim)/2) - np.log((nu+dim)/2)
            return f

        if func0(1e300) >= 0:
            nu = np.inf
        else:
            nu = optimize.bisect(func0, 1e-300, 1e300)
        return nu

    data = data.T
    (dim,n) = data.shape
    mu = np.array([np.median(data,1)]).T
    Sigma = np.cov(data)*(n-1)/n + (1/n)*np.diag(np.var(data, axis=1))
    nu = 20

    last_nu = 0
    i = 0
    while np.abs(last_nu - nu) > tolerance and i < max_iter:
        i += 1
        diffs = data - mu
        delta_iobs = np.sum(diffs * np.linalg.solve(Sigma,diffs), 0)
        
        # update nu
        last_nu = nu
        nu = opt_nu(delta_iobs, nu)
        if nu == np.inf:
            return mu.T[0], Sigma, nu

        # update Sigma
        w_iobs = (nu + dim) / (nu + delta_iobs)
        Sigma = np.dot(w_iobs*diffs, diffs.T) / n

        # update mu
        mu = np.sum(w_iobs * data, 1) / sum(w_iobs)
        mu = np.array([mu]).T

    if i == max_iter:
        print("Warning: EM algorithm did not converge.")
        print("Last nu: ", last_nu)
        print("Current nu: ", nu)

    return mu.T[0], Sigma, nu



from pathlib import Path
from typing import Union

import dill
import numpy as np
from multiprocess import Pool
import torch

from .mcmc import preconditioned_pcn, preconditioned_rwm, pcn, rwm
from .tools import systematic_resample, FunctionWrapper, numpy_to_torch, torch_to_numpy, trim_weights, ProgressBar, flow_numpy_wrapper, effective_sample_size, unique_sample_size
from .scaler import Reparameterize
from .flow import Flow
from .particles import Particles
from .geometry import Geometry
from .threading import configure_threads

class Sampler:
    r"""Preconditioned Monte Carlo class.

    Parameters
    ----------
    prior : callable
        Class implementing the prior distribution.
    likelihood : callable
        Function returning the log likelihood of a set of parameters.
    n_dim : int
        The total number of parameters/dimensions (Optional as it can be infered from the prior class).
    n_effective : int
        The number of effective particles (default is ``n_effective=512``). Higher values
        lead to more accurate results but also increase the computational cost.  This should be
        set to a value that is large enough to ensure that the target distribution is well
        represented by the particles. The number of effective particles should be greater than
        the number of active particles. If ``n_effective=None``, the default value is ``n_effective=2*n_active``.
    n_active : int
        The number of active particles (default is ``n_active=256``). It must be smaller than ``n_effective``.
        For best results, the number of active particles should be no more than half the number of effective particles.
        This is the number of particles that are evolved using MCMC at each iteration. If a pool is provided,
        the number of active particles should be a multiple of the number of processes in the pool to ensure
        efficient parallelisation. If ``n_active=None``, the default value is ``n_active=n_effective//2``.
    likelihood_args : list
        Extra arguments to be passed to likelihood (default is ``likelihood_args=None``). Example:
        ``likelihood_args=[data]``.
    likelihood_kwargs : dict
        Extra arguments to be passed to likelihood (default is ``likelihood_kwargs=None``). Example:
        ``likelihood_kwargs={"data": data}``.
    vectorize : bool
        If True, vectorize ``likelihood`` calculation (default is ``vectorize=False``). If False,
        the likelihood is calculated for each particle individually. If ``vectorize=True``, the likelihood
        is calculated for all particles simultaneously. This can lead to a significant speed-up if the likelihood
        function is computationally expensive. However, it requires that the likelihood function can handle
        arrays of shape ``(n_active, n_dim)`` as input and return an array of shape ``(n_active,)`` as output.
    blobs_dtype : list
        Data type of the blobs returned by the likelihood function (default is ``blobs_dtype=None``). If ``blobs_dtype``
        is not provided, the data type is inferred from the blobs returned by the likelihood function. If the blobs
        are not of the same data type, they are converted to an object array. If the blobs are strings, the data type
        is set to ``object``. If the blobs ``dtype`` is known in advance, it can be provided as a list of data types
        (e.g., ``blobs_dtype=[("blob_1", float), ("blob_2", int)]``). Blobs can be used to store additional data 
        returned by the likelihood function (e.g., chi-squared values, residuals, etc.). Blobs are stored as a
        structured array with named fields when the data type is provided. Currently, the blobs feature is not
        compatible with vectorized likelihood calculations.
    pool : pool or int
        Number of processes to use for parallelisation (default is ``pool=None``). If ``pool`` is an integer
        greater than 1, a ``multiprocessing`` pool is created with the specified number of processes (e.g., ``pool=8``). 
        If ``pool`` is an instance of ``mpi4py.futures.MPIPoolExecutor``, the code runs in parallel using MPI.
        If a pool is provided, the number of active particles should be a multiple of the number of processes in 
        the pool to ensure efficient parallelisation. If ``pool=None``, the code runs in serial mode. When a pool 
        is provided, please ensure that the likelihood function is picklable. 
    pytorch_threads : int
        Maximum number of threads to use for torch. If ``None`` torch uses all
        available threads while training the normalizing flow (default is ``pytorch_threads=1``). 
    flow : ``zuko.flow.Flow`` or str
        Normalizing flow to use for preconditioning (default is ``flow='nsf3'``). Available options are
        ``'nsf3'``, ``'nsf6'``, ``'nsf12'``, ``'maf3'``, ``'maf6'``, and ``'maf12'``. 'nsf' stands for
        Neural Spline Flows and 'maf' stands for Masked Autoregressive Flows. The number indicates the
        number of transformations in the flow. More transformations lead to more flexibility but also
        increase the computational cost. If a ``zuko.flow.Flow`` instance is provided, the normalizing 
        flow is used as is. If a string is provided, a new instance of the normalizing flow is created 
        with the specified architecture. The normalizing flow is used to precondition the MCMC sampler 
        and improve the efficiency of the sampling.
    train_config : dict or ``None``
        Configuration for training the normalizing flow
        (default is ``train_config=None``). Options include a dictionary with the following
        keys: ``"validation_split"``, ``"epochs"``, ``"batch_size"``, ``"patience"``,
        ``"learning_rate"``, ``"annealing"``, ``"gaussian_scale"``, ``"laplace_scale"``,
        ``"noise"``, ``"shuffle"``, ``"clip_grad_norm"``.
    train_frequency : int or None
        Frequency of training the normalizing flow (default is ``train_frequency=None``).
        If ``train_frequency=None``, the normalizing flow is trained every ``n_effective//n_active``
        iterations. If ``train_frequency=1``, the normalizing flow is trained at every iteration.
        If ``train_frequency>1``, the normalizing flow is trained every ``train_frequency`` iterations.
    precondition : bool
        If True, use preconditioned MCMC (default is ``precondition=True``). If False,
        use standard MCMC without normalizing flow. The use of preconditioned MCMC is
        recommended as it is more efficient and scales better with the number of parameters. 
        However, it requires the use of a normalizing flow and the training of the flow
        can be computationally expensive. If ``precondition=False``, the normalizing flow
        is not used and the sampler runs in standard mode. This works well for targets that
        are not multimodal or have strong non-linear correlations between parameters.
    dynamic : bool
        If True, dynamically adjust the effective sample size (ESS) threshold based on the
        number of unique particles (default is ``dynamic=False``). This can be useful for
        targets with a large number of modes or strong non-linear correlations between parameters.
    metric : str
        Metric used for determining the next temperature (``beta``) level (default is ``metric="ess"``).
        Options are ``"ess"`` (Effective Sample Size) or ``"uss"`` (Unique Sample Size). The metric
        is used to determine the next temperature level based on the ESS or USS of the importance
        weights. If the ESS or USS of the importance weights is below the target threshold, the temperature
        is increased. If the ESS or USS is above the target threshold, the temperature is decreased. The
        target threshold is set by the ``n_effective`` parameter.
    n_prior : int
        Number of prior samples to draw (default is ``n_prior=2*(n_effective//n_active)*n_active``). This
        is used to initialise the particles at the beginning of the run. The prior samples are used to
        warm-up the sampler and ensure that the particles are well distributed across the prior volume.
    sample : ``str``
        Type of MCMC sampler to use (default is ``sample="tpcn"``). Options are
        ``"pcn"`` (t-preconditioned Crank-Nicolson) or ``"rwm"`` (Random-walk Metropolis).
        t-preconditioned Crank-Nicolson is the default and recommended sampler for PMC as it
        is more efficient and scales better with the number of parameters.
    n_steps : int
        Number of MCMC steps after logP plateau (default is ``n_steps=n_dim``). This is used
        for early stopping of MCMC. Higher values can lead to better exploration but also
        increase the computational cost. If ``n_steps=None``, the default value is ``n_steps=n_dim``.
    n_max_steps : int
        Maximum number of MCMC steps (default is ``n_max_steps=10*n_dim``).
    resample : ``str``
        Resampling scheme to use (default is ``resample="mult"``). Options are
        ``"syst"`` (systematic resampling) or ``"mult"`` (multinomial resampling).
    output_dir : ``str`` or ``None``
        Output directory for storing the state files of the
        sampler. Default is ``None`` which creates a ``states``
        directory. Output files can be used to resume a run.
    output_label : ``str`` or ``None``
        Label used in state files. Defaullt is ``None`` which
        corresponds to ``"pmc"``. The saved states are named
        as ``"{output_dir}/{output_label}_{i}.state"`` where
        ``i`` is the iteration index.  Output files can be
        used to resume a run.
    random_state : int or ``None``
        Initial random seed.
    """

    def __init__(self,
                 prior: callable,
                 likelihood: callable,
                 n_dim: int = None,
                 n_effective: int = 512,
                 n_active: int = 256,
                 likelihood_args: list = None,
                 likelihood_kwargs: dict = None,
                 vectorize: bool = False,
                 blobs_dtype: str = None,
                 pool=None,
                 pytorch_threads=1,
                 flow='nsf3',
                 train_config: dict = None,
                 train_frequency: int = None,
                 precondition: bool = True,
                 dynamic: bool = True,
                 metric: str = 'ess',
                 n_prior: int = None,
                 sample: str = 'tpcn',
                 n_steps: int = None,
                 n_max_steps: int = None,
                 resample: str = 'mult',
                 output_dir: str = None,
                 output_label: str = None,
                 random_state: int = None,
                 # deprecated
                 n_ess: int = None,
                 ):
        
        # Deprecation warnings
        if n_ess is not None:
            n_effective = n_ess
            # raise warning and print it but do not raise exception
            import warnings
            warnings.warn("n_ess is deprecated. Use n_effective instead.", DeprecationWarning, stacklevel=2)

        # Random seed
        if random_state is not None:
            np.random.seed(random_state)
            torch.manual_seed(random_state)
        self.random_state = random_state

        # Configure PyTorch threads
        configure_threads(pytorch_threads=pytorch_threads)

        # Prior distribution
        self.prior = prior
        self.log_prior = self.prior.logpdf
        self.sample_prior = self.prior.rvs
        self.bounds = self.prior.bounds

        # Log likelihood function
        self.log_likelihood = FunctionWrapper(
            likelihood,
            likelihood_args,
            likelihood_kwargs
        )

        # Blobs data type
        self.blobs_dtype = blobs_dtype
        self.have_blobs = blobs_dtype is not None

        # Number of parameters
        if n_dim is None:
            self.n_dim = self.prior.dim
        else:
            self.n_dim = int(n_dim)

        # Check that at least one parameter is provided
        if n_active is None and n_effective is None:
            raise ValueError("At least one of n_active or n_effective must be provided.")

        # Number of active particles
        if n_active is None:
            self.n_active = int(n_effective/2)
        else:
            self.n_active = int(n_active)

        # Effective Sample Size
        if n_effective is None:
            self.n_effective = int(2*n_active)
        else:
            self.n_effective = int(n_effective)

        # Number of MCMC steps after logP plateau
        if n_steps is None:
            self.n_steps = int(self.n_dim//2)
        else:
            self.n_steps = int(n_steps)

        # Maximum number of MCMC steps
        if n_max_steps is None:
            self.n_max_steps = 10 * self.n_steps
        else:
            self.n_max_steps = int(n_max_steps)

        # Total ESS for termination
        self.n_total = None

        # Particle manager
        self.particles = Particles(n_active, n_dim)

        # Iteration counter
        self.t = 0

        # Parallelism
        self.pool = pool
        if pool is None:
            self.distribute = map
        elif isinstance(pool, int) and pool > 1:
            self.pool = Pool(pool)
            self.distribute = self.pool.map
        else:
            self.distribute = pool.map

        # Likelihood vectorization
        self.vectorize = vectorize
        if self.vectorize and self.have_blobs:
            raise ValueError("Cannot vectorize likelihood with blobs.")

        # Geometry
        self.u_geometry = Geometry()
        self.theta_geometry = Geometry()

        # Normalizing Flow
        self.flow = Flow(self.n_dim, flow)
        self.train_config = dict(validation_split=0.5,
                                 epochs=5000,
                                 batch_size=np.minimum(self.n_effective//2, 512),
                                 patience=int(self.n_dim),
                                 learning_rate=1e-3,
                                 annealing=False,
                                 gaussian_scale=None,
                                 laplace_scale=None,
                                 noise=None,
                                 shuffle=True,
                                 clip_grad_norm=1.0,
                                 verbose=0,
                                )
        if train_config is not None:
            for key in train_config.keys():
                self.train_config[key] = train_config[key]
        
        if train_frequency is None:
            self.train_frequency = np.maximum(self.n_effective//(self.n_active*2), 1)
        else:
            self.train_frequency = int(train_frequency)

        self.flow_untrained = True

        # Scaler
        self.scaler = Reparameterize(self.n_dim, bounds=self.bounds)

        # Output
        if output_dir is None:
            self.output_dir = Path("states")
        else:
            self.output_dir = output_dir
        if output_label is None:
            self.output_label = "pmc"
        else:
            self.output_label = output_label

        # Normalizing flow preconditioning
        self.preconditioned = precondition

        # Effective vs Unique Sample Size
        if metric not in ['ess', 'uss']:
            raise ValueError(f"Invalid metric {metric}. Options are 'ess' or 'uss'.")
        else:
            self.metric = metric

        # Dynamic ESS
        self.dynamic = dynamic
        self.dynamic_ratio = unique_sample_size(np.ones(self.n_effective), k=self.n_active) / self.n_active

        # Sampling algorithm
        if sample not in ['tpcn', 'rwm']:
            raise ValueError(f"Invalid sample {sample}. Options are 'tpcn' or 'rwm'.")
        else:
            self.sample = sample

        # Proposal scale
        self.proposal_scale = 2.38 / self.n_dim ** 0.5

        # Resampling algorithm
        if resample not in ['mult', 'syst']:
            raise ValueError(f"Invalid resample {resample}. Options are 'mult' or 'syst'.")
        else:
            self.resample = resample

        # Prior samples to draw
        if n_prior is None:
            self.n_prior = int(2 * np.maximum(self.n_effective//self.n_active, 1) * self.n_active)
        else:
            self.n_prior = int(np.maximum(n_prior/self.n_active, 1) * self.n_active)
        self.prior_samples = None

        self.logz = None
        self.logz_err = None

        self.current_particles = None
        self.warmup = True
        self.calls = 0
        
        self.progress = None
        self.pbar = None

    def run(self,
            n_total: int = 4096,
            n_evidence: int = 4096,
            progress: bool = True,
            resume_state_path: Union[str, Path] = None,
            save_every: int = None):
        r"""Run Preconditioned Monte Carlo.

        Parameters
        ----------
        n_total : int
            The total number of effectively independent samples to be
            collected (default is ``n_total=2048``).
        n_evidence : int
            The number of importance samples used to estimate the
            evidence (default is ``n_evidence=4096``). If ``n_evidence=0``,
            the evidence is not estimated using importance sampling and the
            SMC estimate is used instead. If ``preconditioned=False``, 
            the evidence is estimated using SMC and ``n_evidence`` is ignored.
        progress : bool
            If True, print progress bar (default is ``progress=True``).
        resume_state_path : ``Union[str, Path]``
            Path of state file used to resume a run. Default is ``None`` in which case
            the sampler does not load any previously saved states. An example of using
            this option to resume or continue a run is e.g. ``resume_state_path = "states/pmc_1.state"``.
        save_every : ``int`` or ``None``
            Argument which determines how often (i.e. every how many iterations) ``pocoMC`` saves
            state files to the ``output_dir`` directory. Default is ``None`` in which case no state
            files are stored during the run.
        """
        if resume_state_path is not None:
            self.load_state(resume_state_path)
            t0 = self.t
            # Initialise progress bar
            self.pbar = ProgressBar(self.progress)
            self.pbar.update_stats(dict(calls=self.particles.get("calls", -1),
                                        beta=self.particles.get("beta", -1),
                                        logZ=self.particles.get("logz", -1)))
        else:
            t0 = self.t
            # Run parameters
            self.n_total = int(n_total)
            self.progress = progress

            # Initialise progress bar
            self.pbar = ProgressBar(self.progress)

        # Initialise particles
        if self.prior_samples is None:
            self.prior_samples = self.sample_prior(self.n_prior)
            self.scaler.fit(self.prior_samples)

        if self.warmup:
            for i in range(self.n_prior//self.n_active):
                if save_every is not None:
                    if (self.t - t0) % int(save_every) == 0 and self.t != t0:
                        self.save_state(Path(self.output_dir) / f'{self.output_label}_{self.t}.state')
                # Set state parameters
                x = self.prior_samples[i*self.n_active:(i+1)*self.n_active]
                u = self.scaler.forward(x)
                logdetj = self.scaler.inverse(u)[1]
                logp = self.log_prior(x)
                logl, blobs = self._log_like(x)
                self.calls += self.n_active

                self.current_particles = dict(u=u,x=x,logl=logl,logp=logp,logdetj=logdetj,
                                    logw=-1e300 * np.ones(self.n_active), blobs=blobs, iter=self.t,
                                    calls=self.calls, steps=1, efficiency=1.0, ess=self.n_effective, 
                                    accept=1.0, beta=0.0, logz=0.0)
                
                self.particles.update(self.current_particles)

                self.pbar.update_stats(dict(calls=self.particles.get("calls", -1), 
                                            beta=self.particles.get("beta", -1), 
                                            logZ=self.particles.get("logz", -1),
                                            ESS=self.particles.get("ess", -1),
                                            acc=self.particles.get("accept", -1),
                                            steps=self.particles.get("steps", -1),
                                            logP=np.mean(self.particles.get("logp", -1)+self.particles.get("logl", -1)),
                                            eff=self.particles.get("efficiency", -1)))
                
                self.pbar.update_iter()

                self.t += 1
            self.warmup = False

        # Run Sequential Monte Carlo
        while self._not_termination(self.current_particles):
            if save_every is not None:
                if (self.t - t0) % int(save_every) == 0 and self.t != t0:
                    self.save_state(Path(self.output_dir) / f'{self.output_label}_{self.t}.state')

            # Choose next beta based on ESS of weights
            self.current_particles = self._reweight(self.current_particles)

            # Train Preconditioner
            self.current_particles = self._train(self.current_particles)

            # Resample particles
            self.current_particles = self._resample(self.current_particles)

            # Evolve particles using MCMC
            self.current_particles = self._mutate(self.current_particles)   

            # Save particles
            self.particles.update(self.current_particles)

        # Compute evidence
        if n_evidence > 0 and self.preconditioned:
            self._compute_evidence(int(n_evidence))
        else:
            _, self.logz = self.particles.compute_logw_and_logz(1.0)
            self.logz_err = None

        self.pbar.close()

    def _not_termination(self, current_particles):
        """
        Check if termination criterion is satisfied.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.
        
        Returns
        -------
        termination : bool
            True if termination criterion is not satisfied.
        """
        log_weights, _ = self.particles.compute_logw_and_logz(1.0)
        weights = np.exp(log_weights - np.max(log_weights))
        if self.metric == 'ess':
            ess = effective_sample_size(weights)
        elif self.metric == 'uss':
            ess = unique_sample_size(weights)

        return 1.0 - current_particles.get("beta") >= 1e-4 or ess < self.n_total

    
    def _mutate(self, current_particles):
        """
        Evolve particles using MCMC.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.
        
        Returns
        -------
        current_particles : dict
            Dictionary containing the updated particles.
        """
        if self.have_blobs:
            blobs = current_particles.get("blobs").copy()
        else:
            blobs = None
        state_dict = dict(
            u=current_particles.get("u").copy(),
            x=current_particles.get("x").copy(),
            logdetj=current_particles.get("logdetj").copy(),
            logp=current_particles.get("logp").copy(),
            logl=current_particles.get("logl").copy(),
            beta=current_particles.get("beta"),
            blobs=blobs,
        )

        function_dict = dict(
            loglike=self._log_like,
            logprior=self.log_prior,
            scaler=self.scaler,
            flow=self.flow,
            u_geometry=self.u_geometry,
            theta_geometry=self.theta_geometry,
        )

        option_dict = dict(
            n_max=self.n_max_steps,
            n_steps=self.n_steps,
            progress_bar=self.pbar,
            proposal_scale=self.proposal_scale,
        )

        if self.preconditioned and self.sample == "tpcn":
            results = preconditioned_pcn(
                state_dict,
                function_dict,
                option_dict
                )
        elif self.preconditioned and self.sample == "rwm":
            results = preconditioned_rwm(
                state_dict,
                function_dict,
                option_dict
                )
        elif not self.preconditioned and self.sample == "tpcn":
            results = pcn(
                state_dict,
                function_dict,
                option_dict
                )
        elif not self.preconditioned and self.sample == "rwm":
            results = rwm(
                state_dict,
                function_dict,
                option_dict
                )

        current_particles["u"] = results.get('u').copy()
        current_particles["x"] = results.get('x').copy()
        current_particles["logdetj"] = results.get('logdetj').copy()
        current_particles["logl"] = results.get('logl').copy()
        current_particles["logp"] = results.get('logp').copy()
        if self.have_blobs:
            current_particles["blobs"] = results.get('blobs').copy()
        current_particles["efficiency"] = results.get('efficiency') / (2.38 / self.n_dim ** 0.5)
        current_particles["steps"] = results.get('steps')
        current_particles["accept"] = results.get('accept')
        current_particles["calls"] = current_particles.get("calls") + results.get('calls')
        self.proposal_scale = results.get('proposal_scale')

        return current_particles


    def _train(self, current_particles):
        """
        Train normalizing flow.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.
        
        Returns
        -------
        current_particles : dict
            Dictionary containing the updated particles.
        """
        u = current_particles.get("u")
        w = current_particles.get("weights")

        if self.preconditioned and (self.t % self.train_frequency == 0 or current_particles.get("beta")==1.0 or self.flow_untrained):
            self.flow_untrained = False
            self.flow.fit(numpy_to_torch(u),
                          weights=numpy_to_torch(w),
                          validation_split=self.train_config["validation_split"],
                          epochs=self.train_config["epochs"],
                          batch_size=int(np.minimum(len(u)//2, self.train_config["batch_size"])),
                          gaussian_scale=self.train_config["gaussian_scale"],
                          laplace_scale=self.train_config["laplace_scale"],
                          patience=self.train_config["patience"],
                          learning_rate=self.train_config["learning_rate"],
                          annealing=self.train_config["annealing"],
                          noise=self.train_config["noise"],
                          shuffle=self.train_config["shuffle"],
                          clip_grad_norm=self.train_config["clip_grad_norm"],
                          verbose=self.train_config["verbose"],
                          )
            
            theta = flow_numpy_wrapper(self.flow).forward(u)[0]
            self.theta_geometry.fit(theta, weights=w)
        else:
            self.u_geometry.fit(u, weights=w)



        return current_particles

    def _resample(self, current_particles):
        """
        Resample particles.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.

        Returns
        -------
        current_particles : dict
            Dictionary containing the updated particles.
        """
        u = current_particles.get("u")
        x = current_particles.get("x")
        logdetj = current_particles.get("logdetj")
        logl = current_particles.get("logl")
        logp = current_particles.get("logp")
        weights = current_particles.get("weights")
        blobs = current_particles.get("blobs")

        if self.resample == 'mult':
            idx_resampled = np.random.choice(np.arange(len(weights)), size=self.n_active, replace=True, p=weights)
        elif self.resample == 'syst':
            idx_resampled = systematic_resample(self.n_active, weights=weights)

        current_particles["u"] = u[idx_resampled]
        current_particles["x"] = x[idx_resampled]
        current_particles["logdetj"] = logdetj[idx_resampled]
        current_particles["logl"] = logl[idx_resampled]
        current_particles["logp"] = logp[idx_resampled]
        if self.have_blobs:
            current_particles["blobs"] = blobs[idx_resampled]
    
        return current_particles

    def _reweight(self, current_particles):
        """
        Reweight particles.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.

        Returns
        -------
        current_particles : dict
            Dictionary containing the updated particles.
        """
        # Update iteration index
        self.t += 1
        self.pbar.update_iter()

        beta_prev = self.particles.get("beta", index=-1)
        beta_max = 1.0
        beta_min = np.copy(beta_prev)

        def get_weights_and_ess(beta):
            logw, _ = self.particles.compute_logw_and_logz(beta)
            weights = np.exp(logw - np.max(logw))
            if self.metric == 'ess':
                ess_est = effective_sample_size(weights)
            elif self.metric == 'uss':
                ess_est = unique_sample_size(weights)
            return weights, ess_est

        weights_prev, ess_est_prev = get_weights_and_ess(beta_prev)
        weights_max, ess_est_max = get_weights_and_ess(beta_max)


        if ess_est_prev <= self.n_effective:
            beta = beta_prev
            weights = weights_prev
            logz = self.particles.get("logz", index=-1)
            ess_est = ess_est_prev
            self.pbar.update_stats(dict(beta=beta, ESS=ess_est_prev, logZ=logz))
        elif ess_est_max >= self.n_effective:
            beta = beta_max 
            weights = weights_max
            _, logz = self.particles.compute_logw_and_logz(beta)
            ess_est = ess_est_max
            self.pbar.update_stats(dict(beta=beta, ESS=ess_est_max, logZ=logz))
        else:
            while True:
                beta = (beta_max + beta_min) * 0.5

                weights, ess_est = get_weights_and_ess(beta)

                if np.abs(ess_est - self.n_effective) < 0.01 * self.n_effective or beta == 1.0:
                    _, logz = self.particles.compute_logw_and_logz(beta)
                    self.pbar.update_stats(dict(beta=beta, ESS=ess_est, logZ=logz))
                    break
                elif ess_est < self.n_effective:
                    beta_max = beta
                else:
                    beta_min = beta

        logw, _ = self.particles.compute_logw_and_logz(beta)
        weights = np.exp(logw - np.max(logw))
        weights /= np.sum(weights)

        if self.dynamic:
            # Adjust the number of effective particles based on the expected number of unique particles
            n_unique_active = unique_sample_size(weights, k=self.n_active)
            # Maintain the original ratio of unique active to effective particles
            if n_unique_active < self.n_active * (0.95 * self.dynamic_ratio):
                self.n_effective = int(self.n_active/n_unique_active * self.n_effective)
            elif n_unique_active > self.n_active * np.minimum(1.05 * self.dynamic_ratio, 1.0):
                self.n_effective = int(n_unique_active/self.n_active * self.n_effective)

        idx, weights = trim_weights(np.arange(len(weights)), weights, ess=0.99, bins=1000)
        current_particles["u"] = self.particles.get("u", index=None, flat=True)[idx]
        current_particles["x"] = self.particles.get("x", index=None, flat=True)[idx]
        current_particles["logdetj"] = self.particles.get("logdetj", index=None, flat=True)[idx]
        current_particles["logl"] = self.particles.get("logl", index=None, flat=True)[idx]
        current_particles["logp"] = self.particles.get("logp", index=None, flat=True)[idx]
        if self.have_blobs:
            current_particles["blobs"] = self.particles.get("blobs", index=None, flat=True)[idx]
        current_particles["logz"] = logz
        current_particles["beta"] = beta
        current_particles["weights"] = weights
        current_particles["ess"] = ess_est

        return current_particles

    def _log_like(self, x):
        """
        Compute log likelihood.

        Parameters
        ----------
        x : array_like
            Array of parameter values.
        
        Returns
        -------
        logl : float
            Log likelihood.
        blob : array_like
            Additional data (default is ``None``).
        """
        if self.vectorize:
            return self.log_likelihood(x), None
        elif self.pool is not None:
            results = list(self.distribute(self.log_likelihood, x))
        else:
            results = list(map(self.log_likelihood, x))


        try:
            blob = [l[1:] for l in results if len(l) > 1]
            if not len(blob):
                raise IndexError
            logl = np.array([float(l[0]) for l in results])
            self.have_blobs = True
        except (IndexError, TypeError):
            logl = np.array([float(l) for l in results])
            blob = None
        else:
            # Get the blobs dtype
            if self.blobs_dtype is not None:
                dt = self.blobs_dtype
            else:
                try:
                    dt = np.atleast_1d(blob[0]).dtype
                except ValueError:
                    dt = np.dtype("object")
                if dt.kind in "US":
                    # Strings need to be object arrays or we risk truncation
                    dt = np.dtype("object")
            blob = np.array(blob, dtype=dt)

            # Deal with single blobs properly
            shape = blob.shape[1:]
            if len(shape):
                axes = np.arange(len(shape))[np.array(shape) == 1] + 1
                if len(axes):
                    blob = np.squeeze(blob, tuple(axes))

        return logl, blob
        
    def evidence(self):
        """
        Return the log evidence estimate and error.
        """
        return self.logz, self.logz_err
        
    def _compute_evidence(self, n=5_000):
        """
        Estimate the evidence using importance sampling.

        Parameters
        ----------
        n : int
            Number of importance samples (default is ``n=5_000``).
        
        Returns
        -------
        logz : float
            Estimate of the log evidence.
        dlogz : float
            Estimate of the error on the log evidence.
        """
        with torch.no_grad():
            theta_q, logq = self.flow.sample(n)
            theta_q = torch_to_numpy(theta_q)
            logq = torch_to_numpy(logq)

        x_q, logdetj = self.scaler.inverse(theta_q)
        logl, _ = self._log_like(x_q)
        logp = self.log_prior(x_q)

        logw = logl + logp + logdetj - logq 
        logz = np.logaddexp.reduce(logw) - np.log(len(logw))

        dlogz = np.std([np.logaddexp.reduce(logw[np.random.choice(len(logw), len(logw))]) - np.log(len(logw)) for _ in range(np.maximum(n,1000))])

        self.logz = logz
        self.logz_err = dlogz
        return logz, dlogz

    def __getstate__(self):
        """
        Get state information for pickling.
        """
        state = self.__dict__.copy()

        try:
            # remove random module
            # del state['rstate']

            # deal with pool
            if state['pool'] is not None:
                del state['pool']  # remove pool
                del state['distribute']  # remove `pool.map` function hook
        except:  # TODO use specific exception type
            pass

        return state

    def posterior(self, resample=False, return_blobs=False, trim_importance_weights=True, return_logw=False, ess_trim=0.99, bins_trim=1_000):
        """
        Return posterior samples.

        Parameters
        ----------
        resample : bool
            If True, resample particles (default is ``resample=False``).
        trim_importance_weights : bool
            If True, trim importance weights (default is ``trim_importance_weights=True``).
        return_logw : bool
            If True, return log importance weights (default is ``return_logw=False``).
        ess_trim : float
            Effective sample size threshold for trimming (default is ``ess_trim=0.99``).
        bins_trim : int
            Number of bins for trimming (default is ``bins_trim=1_000``).

        Returns
        -------
        samples : ``np.ndarray``
            Samples from the posterior.
        weights : ``np.ndarray``
            Importance weights.
        logl : ``np.ndarray``
            Log likelihoods.
        logp : ``np.ndarray``
            Log priors.
        """
        if return_blobs and not self.have_blobs:
            raise ValueError("No blobs available.")

        samples = self.particles.get("x", flat=True)
        logl = self.particles.get("logl", flat=True)
        logp = self.particles.get("logp", flat=True)
        if return_blobs:
            blobs = self.particles.get("blobs", flat=True)
        logw, _ = self.particles.compute_logw_and_logz(1.0)
        weights = np.exp(logw)

        if trim_importance_weights:
            idx, weights = trim_weights(np.arange(len(samples)), weights, ess=ess_trim, bins=bins_trim)
            samples = samples[idx]
            logl = logl[idx]
            logp = logp[idx]
            logw = logw[idx]

        if resample:
            if self.resample == 'mult':
                idx_resampled = np.random.choice(np.arange(len(weights)), size=len(samples), replace=True, p=weights)
            elif self.resample == 'syst':
                idx_resampled = systematic_resample(len(weights), weights=weights)
            if return_blobs:
                return samples[idx_resampled], logl[idx_resampled], logp[idx_resampled], blobs[idx_resampled]
            else:
                return samples[idx_resampled], logl[idx_resampled], logp[idx_resampled]
            
        else:
            if return_logw:
                if return_blobs:
                    return samples, logw, logl, logp, blobs
                else:
                    return samples, logw, logl, logp
            else:
                if return_blobs:
                    return samples, weights, logl, logp, blobs
                else:
                    return samples, weights, logl, logp

    @property
    def results(self):
        """
        Return results.

        Returns
        -------
        results : dict
            Dictionary containing the results.
        """
        return self.particles.compute_results()

    def save_state(self, path: Union[str, Path]):
        """Save current state of sampler to file.

        Parameters
        ----------
        path : ``Union[str, Path]``
            Path to save state.
        """
        print(f'Saving PMC state to {path}')
        Path(path).parent.mkdir(exist_ok=True)
        with open(path, 'wb') as f:
            state = self.__dict__.copy()
            del state['pbar']  # Cannot be pickled
            try:
                # deal with pool
                if state['pool'] is not None:
                    del state['pool']  # remove pool
                    del state['distribute']  # remove `pool.map` function hook
            except BaseException as e:
                print(e)
            dill.dump(file=f, obj=state)

    def load_state(self, path: Union[str, Path]):
        """Load state of sampler from file.

        Parameters
        ----------
        path : ``Union[str, Path]``
            Path from which to load state.
        """
        with open(path, 'rb') as f:
            state = dill.load(file=f)
        self.__dict__ = {**self.__dict__, **state}




import numpy as np
import torch

from .tools import numpy_to_torch, torch_to_numpy, flow_numpy_wrapper
from .student import fit_mvstud

@torch.no_grad()
def preconditioned_pcn(state_dict: dict,
                       function_dict: dict,
                       option_dict: dict):
    """
    Doubly Preconditioned Crank-Nicolson
    
    Parameters
    ----------
    state_dict : dict
        Dictionary of current state
    function_dict : dict
        Dictionary of functions.
    option_dict : dict
        Dictionary of options.
    
    Returns
    -------
    Results dictionary
    """
    # Likelihood call counter
    n_calls = 0

    # Clone state variables
    u = np.copy(state_dict.get('u'))
    x = np.copy(state_dict.get('x'))
    logdetj = np.copy(state_dict.get('logdetj'))
    logl = np.copy(state_dict.get('logl'))
    logp = np.copy(state_dict.get('logp'))
    beta = state_dict.get('beta')
    blobs = state_dict.get('blobs')
    if blobs is None:
        have_blobs = False
    else:
        have_blobs = True

    # Get functions
    log_like = function_dict.get('loglike')
    log_prior = function_dict.get('logprior')
    scaler = function_dict.get('scaler')
    flow = flow_numpy_wrapper(function_dict.get('flow'))
    geometry = function_dict.get('theta_geometry')

    # Get MCMC options
    n_max = option_dict.get('n_max')
    n_steps = option_dict.get('n_steps')
    progress_bar = option_dict.get('progress_bar')
    sigma = np.minimum(option_dict.get('proposal_scale'), 0.99)

    # Get number of particles and parameters/dimensions
    n_walkers, n_dim = x.shape

    # Transform u to theta
    theta, logdetj_flow = flow.forward(u)


    mu = geometry.t_mean
    cov = geometry.t_cov
    nu = geometry.t_nu

    inv_cov = np.linalg.inv(cov)
    chol_cov = np.linalg.cholesky(cov)

    logp2_val = np.mean(logl + logp)
    cnt = 0

    i = 0
    while True:
        i += 1

        diff = theta - mu
        s = np.empty(n_walkers)
        for k in range(n_walkers):
            s[k] = 1./np.random.gamma((n_dim + nu) / 2, 2.0/(nu + np.dot(diff[k],np.dot(inv_cov,diff[k]))))

        # Propose new points in theta space
        theta_prime = np.empty((n_walkers, n_dim))
        for k in range(n_walkers):
            theta_prime[k] = mu + (1.0 - sigma ** 2.0) ** 0.5 * diff[k] + sigma * np.sqrt(s[k]) * np.dot(chol_cov, np.random.randn(n_dim))      

        # Transform to u space
        u_prime, logdetj_flow_prime = flow.inverse(theta_prime)

        # Transform to x space
        x_prime, logdetj_prime = scaler.inverse(u_prime)

        # Compute finite mask
        finite_mask_logdetj_prime = np.isfinite(logdetj_prime)
        finite_mask_x_prime = np.isfinite(x_prime).all(axis=1)
        finite_mask = finite_mask_logdetj_prime & finite_mask_x_prime

        # Compute log-likelihood, log-prior, and log-posterior
        u_rand = np.random.rand(n_walkers)
        logl_prime = np.empty(n_walkers)
        logp_prime = np.empty(n_walkers)
        if have_blobs:
            blobs_prime = np.empty(n_walkers, dtype=np.dtype((blobs[0].dtype, blobs[0].shape)))
            logl_prime[finite_mask], blobs_prime[finite_mask] = log_like(x_prime[finite_mask])
        else:
            logl_prime[finite_mask], _ = log_like(x_prime[finite_mask])
        logp_prime[finite_mask] = log_prior(x_prime[finite_mask])
        logl_prime[~finite_mask] = -np.inf
        logp_prime[~finite_mask] = -np.inf

        n_calls += np.sum(finite_mask)

        # Compute Metropolis factors
        diff_prime = theta_prime-mu
        A = np.empty(n_walkers)
        B = np.empty(n_walkers)
        for k in range(n_walkers):
            A[k] = -(n_dim+nu)/2*np.log(1+np.dot(diff_prime[k],np.dot(inv_cov,diff_prime[k]))/nu)
            B[k] = -(n_dim+nu)/2*np.log(1+np.dot(diff[k],np.dot(inv_cov,diff[k]))/nu)
        alpha = np.minimum(
            np.ones(n_walkers),
            np.exp(logl_prime * beta - logl * beta + logp_prime - logp + logdetj_prime - logdetj + logdetj_flow_prime - logdetj_flow - A + B)
        )
        alpha[np.isnan(alpha)] = 0.0

        # Metropolis criterion
        mask = u_rand < alpha

        # Accept new points
        theta[mask] = theta_prime[mask]
        u[mask] = u_prime[mask]
        x[mask] = x_prime[mask]
        logdetj[mask] = logdetj_prime[mask]
        logdetj_flow[mask] = logdetj_flow_prime[mask]
        logl[mask] = logl_prime[mask]
        logp[mask] = logp_prime[mask]
        if have_blobs:
            blobs[mask] = blobs_prime[mask]

        # Adapt scale parameter using diminishing adaptation
        sigma = np.abs(np.minimum(sigma + 1 / (i + 1)**0.75 * (np.mean(alpha) - 0.234), np.minimum(2.38 / n_dim**0.5, 0.99)))
        #sigma = np.minimum(sigma + 1 / (i + 1)**0.5 * (np.mean(alpha) - 0.234), 0.99)

        # Adapt mean parameter using diminishing adaptation
        mu = mu + 1.0 / (i + 1.0) * (np.mean(theta, axis=0) - mu)

        # Update progress bar if available
        if progress_bar is not None:
            progress_bar.update_stats(
                dict(calls=progress_bar.info['calls'] + np.sum(finite_mask),
                    acc=np.mean(alpha),
                    steps=i,
                    logP=np.mean(logl + logp),
                    eff=sigma / (2.38 / np.sqrt(n_dim)),
                    )
            )

        # Loop termination criteria:
        logp2_val_new = np.mean(logl + logp)
        if logp2_val_new > logp2_val:
            cnt = 0
            logp2_val = logp2_val_new
        else:
            cnt += 1
            if cnt >= n_steps * ((2.38 / n_dim**0.5) / sigma)**2.0:
                break

        if i >= n_max:
            break

    return dict(u=u, x=x, logdetj=logdetj, logl=logl, logp=logp, blobs=blobs, efficiency=sigma, 
                accept=np.mean(alpha), steps=i, calls=n_calls, proposal_scale=sigma)

@torch.no_grad()
def preconditioned_rwm(state_dict: dict,
                       function_dict: dict,
                       option_dict: dict):
    """
    Preconditioned Random-walk Metropolis
    
    Parameters
    ----------
    state_dict : dict
        Dictionary of current state
    function_dict : dict
        Dictionary of functions.
    option_dict : dict
        Dictionary of options.
    
    Returns
    -------
    Results dictionary
    """
    # Likelihood call counter
    n_calls = 0

    # Clone state variables
    u = np.copy(state_dict.get('u'))
    x = np.copy(state_dict.get('x'))
    logdetj = np.copy(state_dict.get('logdetj'))
    logl = np.copy(state_dict.get('logl'))
    logp = np.copy(state_dict.get('logp'))
    beta = state_dict.get('beta')
    blobs = state_dict.get('blobs')
    if blobs is None:
        have_blobs = False
    else:
        have_blobs = True

    # Get functions
    log_like = function_dict.get('loglike')
    log_prior = function_dict.get('logprior')
    scaler = function_dict.get('scaler')
    flow = flow_numpy_wrapper(function_dict.get('flow'))
    geometry = function_dict.get('theta_geometry')

    # Get MCMC options
    n_max = option_dict.get('n_max')
    n_steps = option_dict.get('n_steps')
    progress_bar = option_dict.get('progress_bar')
    sigma = option_dict.get('proposal_scale')

    # Get number of particles and parameters/dimensions
    n_walkers, n_dim = x.shape


    cov = geometry.normal_cov
    chol = np.linalg.cholesky(cov)

    # Transform u to theta
    theta, logdetj_flow = flow.forward(u)

    logp2_val = np.mean(logl + logp + logdetj)
    cnt = 0

    i = 0
    while True:
        i += 1

        # Propose new points in theta space
        theta_prime = np.empty((n_walkers, n_dim))
        for k in range(n_walkers):
            theta_prime[k] = theta[k] + sigma * np.dot(chol, np.random.randn(n_dim))

        # Transform to u space
        u_prime, logdetj_flow_prime = flow.inverse(theta_prime)

        # Transform to x space
        x_prime, logdetj_prime = scaler.inverse(u_prime)

        # Compute finite mask
        finite_mask_logdetj_prime = np.isfinite(logdetj_prime)
        finite_mask_x_prime = np.isfinite(x_prime).all(axis=1)
        finite_mask = finite_mask_logdetj_prime & finite_mask_x_prime

        # Compute log-likelihood, log-prior, and log-posterior
        u_rand = np.random.rand(n_walkers)
        logl_prime = np.empty(n_walkers)
        logp_prime = np.empty(n_walkers)
        if have_blobs:
            blobs_prime = np.empty(n_walkers, dtype=np.dtype((blobs[0].dtype, blobs[0].shape)))
            logl_prime[finite_mask], blobs_prime[finite_mask] = log_like(x_prime[finite_mask])
        else:
            logl_prime[finite_mask], _ = log_like(x_prime[finite_mask])
        logp_prime[finite_mask] = log_prior(x_prime[finite_mask])
        logl_prime[~finite_mask] = -np.inf
        logp_prime[~finite_mask] = -np.inf

        n_calls += np.sum(finite_mask)

        # Compute Metropolis factors
        alpha = np.minimum(
            np.ones(n_walkers),
            np.exp(logl_prime * beta - logl * beta + logp_prime - logp + logdetj_prime - logdetj + logdetj_flow_prime - logdetj_flow)
        )
        alpha[np.isnan(alpha)] = 0.0

        # Metropolis criterion
        mask = u_rand < alpha

        # Accept new points
        theta[mask] = theta_prime[mask]
        u[mask] = u_prime[mask]
        x[mask] = x_prime[mask]
        logdetj[mask] = logdetj_prime[mask]
        logdetj_flow[mask] = logdetj_flow_prime[mask]
        logl[mask] = logl_prime[mask]
        logp[mask] = logp_prime[mask]
        if have_blobs:
            blobs[mask] = blobs_prime[mask]

        # Adapt scale parameter using diminishing adaptation
        sigma = sigma + 1 / (i + 1) * (np.mean(alpha) - 0.234)

        # Update progress bar if available
        if progress_bar is not None:
            progress_bar.update_stats(
                dict(calls=progress_bar.info['calls'] + np.sum(finite_mask),
                    acc=np.mean(alpha),
                    steps=i,
                    logP=np.mean(logl + logp),
                    eff=sigma / (2.38 / np.sqrt(n_dim)))
            )

        # Loop termination criteria:
        logp2_val_new = np.mean(logl + logp + logdetj)
        if logp2_val_new > logp2_val:
            cnt = 0
            logp2_val = logp2_val_new
        else:
            cnt += 1
            if cnt >= n_steps * (np.minimum(1.0, (2.38 / n_dim**0.5) / sigma))**2.0:
                break

        if i >= n_max:
            break


    return dict(u=u, x=x, logdetj=logdetj, logl=logl, logp=logp, blobs=blobs, efficiency=sigma, 
                accept=np.mean(alpha), steps=i, calls=n_calls, proposal_scale=sigma)


def pcn(state_dict: dict,
        function_dict: dict,
        option_dict: dict):
    """
    Preconditioned Crank-Nicolson
    
    Parameters
    ----------
    state_dict : dict
        Dictionary of current state
    function_dict : dict
        Dictionary of functions.
    option_dict : dict
        Dictionary of options.
    
    Returns
    -------
    Results dictionary
    """
    # Likelihood call counter
    n_calls = 0

    # Clone state variables
    u = np.copy(state_dict.get('u'))
    x = np.copy(state_dict.get('x'))
    logdetj = np.copy(state_dict.get('logdetj'))
    logl = np.copy(state_dict.get('logl'))
    logp = np.copy(state_dict.get('logp'))
    beta = state_dict.get('beta')
    blobs = state_dict.get('blobs')
    if blobs is None:
        have_blobs = False
    else:
        have_blobs = True

    # Get functions
    log_like = function_dict.get('loglike')
    log_prior = function_dict.get('logprior')
    scaler = function_dict.get('scaler')
    geometry = function_dict.get('u_geometry')

    # Get MCMC options
    n_max = option_dict.get('n_max')
    n_steps = option_dict.get('n_steps')
    progress_bar = option_dict.get('progress_bar')
    sigma = np.minimum(option_dict.get('proposal_scale'), 0.99)

    # Get number of particles and parameters/dimensions
    n_walkers, n_dim = x.shape

    mu = geometry.t_mean
    cov = geometry.t_cov
    nu = geometry.t_nu

    inv_cov = np.linalg.inv(cov)
    chol_cov = np.linalg.cholesky(cov)

    logp2_val = np.mean(logl + logp)
    #logp2_val = np.mean(logl * beta + logp)
    cnt = 0

    i = 0
    while True:
        i += 1

        diff = u - mu
        s = np.empty(n_walkers)
        for k in range(n_walkers):
            s[k] = 1./np.random.gamma((n_dim + nu) / 2, 2.0/(nu + np.dot(diff[k],np.dot(inv_cov,diff[k]))))

        # Propose new points in u space
        u_prime = np.empty((n_walkers, n_dim))
        for k in range(n_walkers):
            u_prime[k] = mu + (1.0 - sigma ** 2.0) ** 0.5 * diff[k] + sigma * np.sqrt(s[k]) * np.dot(chol_cov, np.random.randn(n_dim))        

        # Transform to x space
        x_prime, logdetj_prime = scaler.inverse(u_prime)

        # Compute finite mask
        finite_mask_logdetj_prime = np.isfinite(logdetj_prime)
        finite_mask_x_prime = np.isfinite(x_prime).all(axis=1)
        finite_mask = finite_mask_logdetj_prime & finite_mask_x_prime

        # Compute log-likelihood, log-prior, and log-posterior
        u_rand = np.random.rand(n_walkers)
        logl_prime = np.empty(n_walkers)
        logp_prime = np.empty(n_walkers)
        if have_blobs:
            blobs_prime = np.empty(n_walkers, dtype=np.dtype((blobs[0].dtype, blobs[0].shape)))
            logl_prime[finite_mask], blobs_prime[finite_mask] = log_like(x_prime[finite_mask])
        else:
            logl_prime[finite_mask], _ = log_like(x_prime[finite_mask])
        logp_prime[finite_mask] = log_prior(x_prime[finite_mask])
        logl_prime[~finite_mask] = -np.inf
        logp_prime[~finite_mask] = -np.inf
        
        n_calls += np.sum(finite_mask)

        # Compute Metropolis factors
        diff_prime = u_prime - mu
        A = np.empty(n_walkers)
        B = np.empty(n_walkers)
        for k in range(n_walkers):
            A[k] = -(n_dim+nu)/2*np.log(1+np.dot(diff_prime[k],np.dot(inv_cov,diff_prime[k]))/nu)
            B[k] = -(n_dim+nu)/2*np.log(1+np.dot(diff[k],np.dot(inv_cov,diff[k]))/nu)
        alpha = np.minimum(
            np.ones(n_walkers),
            np.exp(logl_prime * beta - logl * beta + logp_prime - logp + logdetj_prime - logdetj - A + B)
        )
        alpha[np.isnan(alpha)] = 0.0

        # Metropolis criterion
        mask = u_rand < alpha

        # Accept new points
        u[mask] = u_prime[mask]
        x[mask] = x_prime[mask]
        logdetj[mask] = logdetj_prime[mask]
        logl[mask] = logl_prime[mask]
        logp[mask] = logp_prime[mask]
        if have_blobs:
            blobs[mask] = blobs_prime[mask]

        # Adapt scale parameter using diminishing adaptation
        sigma = np.abs(np.minimum(sigma + 1 / (i + 1)**0.75 * (np.mean(alpha) - 0.234), np.minimum(2.38 / n_dim**0.5, 0.99)))
        #sigma = sigma + 1 / (i + 1)**0.75 * (np.mean(alpha) - 0.234)

        # Update progress bar if available
        if progress_bar is not None:
            progress_bar.update_stats(
                dict(calls=progress_bar.info['calls'] + np.sum(finite_mask),
                    acc=np.mean(alpha),
                    steps=i,
                    logP=np.mean(logl + logp),
                    eff=sigma / (2.38 / np.sqrt(n_dim)))
            )

        # Loop termination criteria:
        logp2_val_new = np.mean(logl + logp)
        if logp2_val_new > logp2_val:
            cnt = 0
            logp2_val = logp2_val_new
        else:
            cnt += 1
            if cnt >= n_steps * ((2.38 / n_dim**0.5) / sigma)**2.0:
                break

        if i >= n_max:
            break

    return dict(u=u, x=x, logdetj=logdetj, logl=logl, logp=logp, blobs=blobs, efficiency=sigma, 
                accept=np.mean(alpha), steps=i, calls=n_calls, proposal_scale=sigma)

def rwm(state_dict: dict,
        function_dict: dict,
        option_dict: dict):
    """
    Random-walk Metropolis
    
    Parameters
    ----------
    state_dict : dict
        Dictionary of current state
    function_dict : dict
        Dictionary of functions.
    option_dict : dict
        Dictionary of options.
    
    Returns
    -------
    Results dictionary
    """
    # Likelihood call counter
    n_calls = 0

    # Clone state variables
    u = np.copy(state_dict.get('u'))
    x = np.copy(state_dict.get('x'))
    logdetj = np.copy(state_dict.get('logdetj'))
    logl = np.copy(state_dict.get('logl'))
    logp = np.copy(state_dict.get('logp'))
    beta = state_dict.get('beta')
    blobs = state_dict.get('blobs')
    if blobs is None:
        have_blobs = False
    else:
        have_blobs = True

    # Get functions
    log_like = function_dict.get('loglike')
    log_prior = function_dict.get('logprior')
    scaler = function_dict.get('scaler')
    geometry = function_dict.get('u_geometry')

    # Get MCMC options
    n_max = option_dict.get('n_max')
    n_steps = option_dict.get('n_steps')
    progress_bar = option_dict.get('progress_bar')
    sigma = option_dict.get('proposal_scale')

    # Get number of particles and parameters/dimensions
    n_walkers, n_dim = x.shape

    cov = geometry.normal_cov
    chol = np.linalg.cholesky(cov)

    logp2_val = np.mean(logl + logp + logdetj)
    cnt = 0

    i = 0
    while True:
        i += 1

        # Propose new points in theta space
        u_prime = np.empty((n_walkers, n_dim))
        for k in range(n_walkers):
            u_prime[k] = u[k] + sigma * np.dot(chol, np.random.randn(n_dim))

        # Transform to x space
        x_prime, logdetj_prime = scaler.inverse(u_prime)

        # Compute finite mask
        finite_mask_logdetj_prime = np.isfinite(logdetj_prime)
        finite_mask_x_prime = np.isfinite(x_prime).all(axis=1)
        finite_mask = finite_mask_logdetj_prime & finite_mask_x_prime

        # Compute log-likelihood, log-prior, and log-posterior
        u_rand = np.random.rand(n_walkers)
        logl_prime = np.empty(n_walkers)
        logp_prime = np.empty(n_walkers)
        if have_blobs:
            blobs_prime = np.empty(n_walkers, dtype=np.dtype((blobs[0].dtype, blobs[0].shape)))
            logl_prime[finite_mask], blobs_prime[finite_mask] = log_like(x_prime[finite_mask])
        else:
            logl_prime[finite_mask], _ = log_like(x_prime[finite_mask])
        logp_prime[finite_mask] = log_prior(x_prime[finite_mask])
        logl_prime[~finite_mask] = -np.inf
        logp_prime[~finite_mask] = -np.inf

        n_calls += np.sum(finite_mask)

        # Compute Metropolis factors
        alpha = np.minimum(
            np.ones(n_walkers),
            np.exp(logl_prime * beta - logl * beta + logp_prime - logp + logdetj_prime - logdetj)
        )
        alpha[np.isnan(alpha)] = 0.0

        # Metropolis criterion
        mask = u_rand < alpha

        # Accept new points
        u[mask] = u_prime[mask]
        x[mask] = x_prime[mask]
        logdetj[mask] = logdetj_prime[mask]
        logl[mask] = logl_prime[mask]
        logp[mask] = logp_prime[mask]
        if have_blobs:
            blobs[mask] = blobs_prime[mask]

        # Adapt scale parameter using diminishing adaptation
        sigma = np.abs(sigma + 1 / (i + 1) * (np.mean(alpha) - 0.234))

        # Update progress bar if available
        if progress_bar is not None:
            progress_bar.update_stats(
                dict(calls=progress_bar.info['calls'] + np.sum(finite_mask),
                    acc=np.mean(alpha),
                    steps=i,
                    logP=np.mean(logl + logp),
                    eff=sigma / (2.38 / np.sqrt(n_dim)))
            )

        # Loop termination criteria:
        logp2_val_new = np.mean(logl + logp + logdetj)
        if logp2_val_new > logp2_val:
            cnt = 0
            logp2_val = logp2_val_new
        else:
            cnt += 1
            if cnt >= n_steps * ((2.38 / n_dim**0.5) / sigma)**2.0:
                break

        if i >= n_max:
            break


    return dict(u=u, x=x, logdetj=logdetj, logl=logl, logp=logp, blobs=blobs, efficiency=sigma, 
                accept=np.mean(alpha), steps=i, calls=n_calls, proposal_scale=sigma)
